
{Convert double precision binary to BCD
Uses conventional method of repeatedly subtracting powers of 10, beginning
with 1E19 and finishing with 1E0.  Subtracts each power of ten until
the residual is negative, does a restoring add, and goes on to next
smaller power of ten.
The inner loop is:

IncBCD  add 0-0, #1                 'because we increment first
SubLo   sub ValLo, 0-0 wc          'subtract low part
SubHi   subx ValHi, 0-0 wc          'hi part
        if_NC jmp #IncBCD           'do it again

so that four instructions are needful; one to increment the BCD digit, two to subtract
and one to determine if the result borrowed}

CON
        _clkmode = xtal1 + pll16x    'Standard clock mode 
        _xinfreq = 5_000_000         '* crystal frequency = 80 MHz
  
VAR                                       
  long  MyParm[10]                'Bin2BCDing addresses, etc to pasm cog
  byte  Cog       
  long  Command                  'command to cog
  long  Binary[2]                'most significant, least sig
  byte  BCD[20]                  'result in bcd
  byte  digit                    'used to display decimal

OBJ
  pst      : "parallax serial terminal"
  
PUB main                                    
    pst.Start (115_200)                        'start up ser terminal
    pst.str(String("hello, world   "))         'runs in cog 1
    pst.NewLine
    waitcnt(clkfreq/10+cnt)

    MyParm[0] := @Command                  'give command address to cog
    cog := cognew(@BCog, @MyParm)           'start up the conversion cog

    Binary[0] := 0                          'demo first 32 powers of 2
    Binary[1] := 1      
    repeat 32
      Bin2BCD
      Binary[1] := Binary[1] << 1

    Binary[0] := 1                          'next 32 powers of 2
    Binary[1] := 0      
    repeat 32
       Bin2BCD
       Binary[0] := Binary[0] << 1

    binary[0] := $FFFF_FFFF                '2^64 minus 1
    binary[1] := $FFFF_FFFF
    Bin2BCD

    binary[0] := $8ac7_2304                 'dec 999...999
    binary[1] := $89e7_ffff
    Bin2BCD   
                
PUB Bin2BCD                                 
    Command := (@Binary << 16) | @BCD      'start up the coversion cog
    repeat while (command <> 0)            'wait on cog
    pst.newline
    pst.hex(Binary[0]>>16,4)               'format taken from SUBX description
    pst.char("_")
    pst.hex(Binary[0],4)
    pst.char(":")
    pst.hex(Binary[1]>>16,4)
    pst.char("_")
    pst.hex(Binary[1],4)
    
    pst.str(string("      "))
    repeat digit from 0 to 19               'all 20 digits
       if ((digit // 3) == 2)               'stuff comma
          pst.char(",")     
       pst.char(bcd[digit] + "0")
 
DAT   '***********conversion cog******************8
BCog    org   0                '               '
        mov AdPar,Par           'get the address of the command long
        rdlong AdCommand, AdPar 'command long
        'or dira, syncMask      'uncomment to see timing on I/O 0
        
loop    rdlong work, AdCommand wz   'see if a command active
        if_Z jmp #loop              'none
        or outa, syncmask           'officially busy         
        mov BCDAdrs, work           'get a copy of addresses
        and BCDAdrs, HubMask        'address of BCD array in hub
        mov HubAdrs, work 
        shr HubAdrs, #16            'scale source address
        and HubAdrs, HubMask        'address of argument
        rdlong ValHi, HubAdrs       'get high or argument
        add HubAdrs, #4             'point to low order
        rdlong ValLo, HubAdrs       'fetch lo order half
        
'**********Here we initialize the modified instruction to point to 10^19                
        movd InitBCD, #CogBCD       'point to beginning of array
        movd IncBCD, #CogBCD        'also the incrementer
        movs SubLo, #E19Lo          'the subtractors
        movs SubHi, #E19Hi
        movs AddLo, #E19Lo          'the restoring adds
        movs AddHi, #E19Hi               
        mov CogDig, #20            'will do 20 digits
                  
InitBCD neg 0-0, #1                 'set to minus one
IncBCD  add 0-0, #1                 'because we increment once too often
SubLo   sub ValLo, 0-0 wc           'subtract low part
SubHi   subx ValHi, 0-0 wc          'hi part
        if_NC jmp #IncBCD           'do it again if didn't borrow       
AddLo   add ValLo, 0-0 wc           'once too many, restoring add
AddHi   addx ValHi, 0-0

'******Here we go on to the next BCD digit*************
        add InitBCD, NextDest       'point to next bcd digit
        add IncBCD, NextDest
        add SubLo, #1                'point to next power of 10
        add SubHi, #1
        add AddLo, #1
        add AddHi, #1                        
        djnz CogDig, #InitBCD         'do next digit
           
        movd stuff, #CogBCD          'set up to move BCD to hub
        mov CogDig, #20              'all 20 digits
stuff   wrbyte 0-0, BCDAdrs          'move one digit
        add Stuff, NextDest          'cog address
        add BCDAdrs, #1              'hub address
        djnz CogDig, #stuff           '20 times
                                     
        andn outa, syncmask            'finished 
        wrlong  zero,  AdCommand       'release spin cog
        jmp #loop                      'and await more
            
Zero         long  0
NextDest     long  $200          'increment one in dest field
HubMask      long  $7FFF         'mask address into hub
syncMask     long   1             'scope sync on 0
E19Hi        long  2_328_306_436 'Hi portion of 1E19
E18Hi        long  232_830_643
E17Hi        long  23_283_064
E16Hi        long  2_328_306
E15Hi        long  232_830
E14Hi        long  23_283
E13Hi        long  2328
E12Hi        long  232
E11Hi        long  23
E10Hi        long  2
E9Hi         long  0
E8Hi         long  0
E7Hi         long  0
E6Hi         long  0
E5Hi         long  0
E4Hi         long  0
E3Hi         long  0
E2Hi         long  0
E1Hi         long  0
E0Hi         long  0             'Hi portion of 1E0

E19Lo        long  2_313_682_944  'lo portion of 1E19
E18Lo        long  2_808_348_672
E17Lo        long  1_569_325_056
E16Lo        long  1_874_919_424
E15Lo        long  2_764_472_320
E14Lo        long    276_447_232
E13Lo        long  1_316_134_912
E12Lo        long  3_567_587_328
E11Lo        long  1_215_752_192     
E10Lo        long  1_410_065_408
E9Lo         long  1000000000
E8Lo         long  100000000
E7Lo         long  10000000
E6Lo         long  1000000
E5Lo         long  100000
E4Lo         long  10000
E3Lo         long  1000
E2Lo         long  100
E1Lo         long  10
E0Lo         long  1

CogBCD       res   20   'resultant bcd
CogDig       res       'which digit we are doing    
AdPar        res         'parameter list                                                     
AdCommand    res         'command address
Work         res        'generally useful
ValHi        res        'high order half of input argument
ValLo        res        'low order half
HubAdrs      res         'address into hub
BCDAdrs      res         'address where to put result   





                           